# Copyright (c) Alibaba, Inc. and its affiliates.
from .cross_entropy_loss import CrossEntropyLoss
from .iou_loss import IOUloss
from .mse_loss import JointsMSELoss
from .pytorch_metric_learning import *
